(* $Id: authors_base.ml,v 1.5 2000/04/22 16:59:57 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(* Authors_base:
 *
 * A class that is derived from File_base.t and handles the specific
 * issues of the author database.
 *)

(* Attributes of an 'author':
 * "name"       Full name
 * "email"      Email address x@y.z
 * "home_link"  An href pointing to information about the author
 * "password"   The password
 *)

open Common

let scheme =
  Scheme.mk_scheme 
    [ "name",      Field.T_constrained
	             (Field.T_line,
		      (fun f -> if f = Field.null then
			raise(User_error "Field 'Name' must not be empty")));
      "email",     Field.T_line;
      "home_link", Field.T_hyperlink;
      "password",  Field.T_password;
    ] 
    [ "name" ]
    []
;;


(**********************************************************************)
(* Authors_base.t: 
 * Persistent author database. Compared with File_base.t, some checks on
 * the integrity have been added.
 *)

class authors =
  object (self)
    inherit File_base.t scheme (!basedir ^ "/authors.db") "authors" as super


    method check_type fields =
      super # check_type fields;
      let pw = List.assoc "password" fields in
      if pw = Field.null then
	raise(User_error "You must have a password!")

    method add_record fields =
      try
	super # add_record fields 
      with
	Table.Not_unique "name" ->
	  raise (User_error "A user with this name already exists!");


    method change_record id fields =
      try
	super # change_record id fields 
      with
	Table.Not_unique "name" ->
	  raise (User_error "A user with this name already exists!");

  end
;;

(**********************************************************************)

Base.register_table "authors" (new authors :> Table.t);;


(* As we have only ONE author DB, there should be a global variable that
 * contains that DB. The following functions simplify the access to the
 * DB.
 *)

let authenticate name password =
  (* Authenticate an author by name and password. Returns the ID of the
   * author or fails
   *)
  Lock.require_read_permission();
  let db = Base.get_table "authors" in
  let exists r = 
    (Field.mk_line name = Record.field r "name") & 
    (Field.eq_password (Record.field r "password") password)
  in
  let s = db # select exists in
  if s = [] then
    raise (User_error "Login failed!");
  Record.id (List.hd s)
;;


let about_author id =
  (* Get the record for the author with ID 'id' *)
  let db = Base.get_table "authors" in
  db # get_record id
;;

(* ======================================================================
 * History:
 * 
 * $Log: authors_base.ml,v $
 * Revision 1.5  2000/04/22 16:59:57  gerd
 * 	Change: It is avoided to obtain a read lock for the author db
 * only. (Note that also do_protected has changed; it now gets a read lock
 * very early and releases it at the right time.)
 *
 * Revision 1.4  1999/05/13 12:30:52  gerd
 * 	Minor changes in the locking code.
 *
 * Revision 1.3  1999/04/27 20:51:23  gerd
 * 	Change: "authors" and "links" are now tables known in Base. Note
 * that this has the consequence that these tables must be considered
 * as (coerced to) class Table.t, i.e. additional methods are no longer visible.
 *
 * Revision 1.2  1999/04/25 21:38:16  gerd
 * 	Removed some of the functions and put the checks they did
 * directly into the class.
 *
 * Revision 1.1  1999/04/20 18:50:15  gerd
 * 	Moved from ../lib/ to this directory.
 *
 *
 * ======================================================================
 * OLD LOGS:
 *
 * Revision 1.4  1999/04/18 23:14:09  gerd
 * 	Major change: The Db tables have now typed attributes. The
 * attributes are represented as Field.field, grouped in records by
 * Record.record and stored in tables by File_base.t as before. A Scheme.scheme
 * is the signature of a table.
 * 	Change: Layout tries to encapsulate some of the layout properties
 * of tables.
 *
 * Revision 1.3  1999/03/14 01:08:40  gerd
 * 	Added: modification date as new record component
 * 	Added: by default, items are listed by modification date
 * 	Changed: passwords are stored as MD5 message digest
 * 	Added: can now change existing records field by field
 * 	Improved: design of the web pages
 *
 * Revision 1.2  1999/02/21 17:50:00  gerd
 * 	Fixed bug: author database did not get written back to disk after
 * change of login data
 *
 * Revision 1.1  1999/02/18 00:59:00  gerd
 * 	First version, has been tested to some extend.
 *
 * 
 *)
