(* $Id: links_base.ml,v 1.8 2000/04/22 17:00:33 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(* Links_base:
 *
 * A class that is derived from File_base.t and handles the specific
 * issues of the link database.
 * Currently, there are two tables storing links: orig_links and rep_links
 * for originial links resp. reported links.
 *)

open Common
open File_base
open Str
open Record
open Table



(* Fields for both orig_links and rep_links: *)

let common_fields =
  [ "title",                Field.T_constrained
                                (Field.T_line,
				 (fun f -> if f = Field.null then
				   raise(User_error "Field 'title' must not be empty")));
    "subtitle",             Field.T_constrained
	                        (Field.T_line,
				 (fun f -> if f = Field.null then
				   raise(User_error "Field 'subtitle' must not be empty")));
    "owner_id",             Field.T_int;
    "info_link",            Field.T_hyperlink;
    "download_link",        Field.T_hyperlink;
    "modification_date",    Field.T_date;
    "classification",       Field.T_select (Classifications_base.selection);
    "release",              Field.T_line;
    "stability",            Field.T_select (Cats.stability);
    "stable_release",       Field.T_line;
    "license",              Field.T_line;
    "description",          Field.T_textbox;
    "recent_changes",       Field.T_textbox;
    "dep_ocaml",            Field.T_line;
    "dep_os",               Field.T_line;
    "dep_software",         Field.T_textbox;
  ] 
;;


let orig_scheme =
  Scheme.mk_scheme 
    (common_fields @
     [ "author_id",            Field.T_int;
       "author_name",          Field.T_line;
       "author_link",          Field.T_hyperlink;
       "coauthor_name",       Field.T_line;
       "coauthor_link",       Field.T_hyperlink;
     ]) 
    []
    [ "author_id";
      "author_name";
      "author_link"
    ]
;;


let rep_scheme =
  Scheme.mk_scheme
    (common_fields @
     [ "reporter_id",     Field.T_int;
       "reporter_name",   Field.T_line;
       "author_name",     Field.T_line;
       "author_link",     Field.T_hyperlink;
       "coauthor_name",  Field.T_line;
       "coauthor_link",  Field.T_hyperlink;
     ])
    []
    [ "reporter_id";
      "reporter_name" ]
;;


(**********************************************************************)
(* Links_base.t: 
 * Persistent link database. Compared with File_base.t, some checks on
 * the integrity have been added. Furthermore, the derived attribute
 * "author_name" is added.
 *)

class links scheme name =
  object (self)
    inherit File_base.t scheme (!basedir ^ "/" ^ name ^ ".db") name as super


    method get_derived_field r n =
      let lookup attribute id =
	(try
	  let author_id = Field.dest_int id in
	  let author_rec = Authors_base.about_author author_id in
	  field author_rec attribute
	with 
	  Field.No_content -> Field.null)
      in
      match n with
	"author_name"    -> lookup "name" (field r "owner_id")
      |	"reporter_name"  -> lookup "name" (field r "owner_id")
      |	"author_link"    -> lookup "home_link" (field r "owner_id")
      |	"author_id"      -> r # field "owner_id"
      |	"reporter_id"    -> r # field "owner_id"
      |	_ ->
	  failwith ("Links_base.t: unknown derived field " ^ n)

  end
;;

(**********************************************************************)

(* Now the access functions for everybody: *)

Base.register_table 
    "orig_links" 
    (new links orig_scheme "orig_links" :> Table.t)
;;


Base.register_table 
    "rep_links" 
    (new links rep_scheme "rep_links" :> Table.t)
;;


(************ The following currently only for orig_links ****************)


let link_db = 
  lazy begin
    let orig_links = Base.get_table "orig_links" in
    let rep_links  = Base.get_table "rep_links" in
    new View.intersection [orig_links; rep_links]
  end
;;


let release_re = regexp "^[0-9]+\\(\\.[0-9]+\\)$";;
let release_split_re = regexp "\\.";;

let cmp_release a_field b_field =
  (* Compare two release numbers.
   * Returns negative number if b_field is smaller.
   * Return positive number if a_field is smaller.
   * Returns 0 if both releases are identical.
   * Note: This makes that release numbers are sorted in decending order!
   *)
  try
    let a = Field.dest_line a_field in   (* or raise No_content *)
    let b = Field.dest_line b_field in   (* or raise No_content *)
    if string_match release_re a 0 & string_match release_re b 0 then begin
      let a_list = List.map int_of_string (split release_split_re a) in
      let b_list = List.map int_of_string (split release_split_re b) in
      List.fold_left2
	(fun r ak bk -> if r = 0 then -(compare ak bk) else r)
	0
	a_list
	b_list
    end else
      -(compare a b)  (* generic comparison *)
  with
      Field.No_content ->
	(* One or both fields are empty. In this case, the empty field
	 * is smaller than the filled field.
	 *)
	if a_field = Field.null then begin
	  if b_field = Field.null then
	    0
	  else
	    1
	end
	else (-1)
;;


let title_author_links () =
  (* get links sorted first by title, then by release, then by author *)
  let db = Lazy.force link_db in
  db # get_somehow_sorted_records 
    ["title", Field.compare_fields Field.T_line; 
      "release", cmp_release; 
      "author_name", Field.compare_fields Field.T_line]
;;


let author_title_links () =
  (* get links sorted first by author, then by title, then by release *)
  let db = Lazy.force link_db in
  db # get_somehow_sorted_records 
    ["author_name", Field.compare_fields Field.T_line; 
      "title", Field.compare_fields Field.T_line; 
      "release", cmp_release]
;;


let date_author_links () =
  (* get links sorted by date, then by title, then by release, then by author *)
  let db = Lazy.force link_db in
  db # get_somehow_sorted_records 
    [ "modification_date", 
         (fun a b -> 
	   if a = Field.null & b = Field.null then 0 else
	   if a = Field.null then 1 else
	   if b = Field.null then (-1) else
	   Field.compare_fields Field.T_date b a);
      "title", Field.compare_fields Field.T_line; 
      "release", cmp_release;
      "author_name", Field.compare_fields Field.T_line; 
    ]
;;


let classification_author_links () =
  (* get links sorted by classification, then by title, then by release, then by author *)
  let db = Lazy.force link_db in
  db # get_somehow_sorted_records 
    [ "classification", Field.compare_fields 
	                  (Field.T_select (Classifications_base.selection));
      "title", Field.compare_fields Field.T_line; 
      "release", cmp_release;
      "author_name", Field.compare_fields Field.T_line; 
    ]
;;


let classified_links () =
  (* get links sorted by classification; links may be given several times
   * per classification
   *)
  let db = Lazy.force link_db in
  let cl_tab = Base.get_table "classifications" in
  let cl_lst = cl_tab # select (fun _ -> true) in
  let base_lst = title_author_links() in
  let l =
    List.map
      (fun cl ->
	let cl_val = Field.dest_line (cl # field "id") in
	Field.html_output [Field.Field("noname", cl # field "title")],
	List.filter
	  (fun r ->
	     try
	      let f = r # field "classification" in
	      List.mem cl_val (Field.dest_select f)
	    with
	      (Field.No_content|Not_found) -> false)
	  base_lst)
      cl_lst
  in
  Sort.list (fun a b -> String.lowercase(fst a) <= String.lowercase(fst b)) l
;;


let author's_links id =
  (* get all links for the author with ID 'id' sorted first by title,
   * then by release
   *)
  let str_id = Field.mk_int id in
  let db = Lazy.force link_db in
  let recs = db # get_sorted_records ["title"; "release"] in
  List.flatten
    (List.map
       (fun r ->
	 if field r "author_id" = str_id then [ r ] else [])
       recs)
;;



(* TODO: The following functions are deprecated. Remove them. *)

let new_link fields =
  (* add new link to the database *)
  let db = Lazy.force link_db in
  let _ = db # add_record fields in
  db # write_database ()
;;


let change_link id fields =
  (* change link with 'id' to value 'fields' *)
  let db = Lazy.force link_db in
  db # change_record id fields;
  db # write_database ()
;;


let delete_link id =
  (* delete link with ID 'id' *)
  let db = Lazy.force link_db in
  db # delete_record id;
  db # write_database()
;;


let get_link id =
  (* returns record or raises Not_found *)
  let db = Lazy.force link_db in
  db # get_record id
;;


(* ======================================================================
 * History:
 * 
 * $Log: links_base.ml,v $
 * Revision 1.8  2000/04/22 17:00:33  gerd
 * 	Changed the sorting order of the various link lists.
 *
 * Revision 1.7  1999/05/19 19:41:10  gerd
 * 	Bugfix: classified_links works now even if there are records
 * without any classification. BUT: These records do not appear in the
 * result.
 *
 * Revision 1.6  1999/05/18 20:05:25  gerd
 * 	Minor change in the data model: author_link is now for both
 * variants available.
 * 	"link_db" is implemented using the new View.intersection
 * feature.
 * 	Added "classified_links" to get links sorted by classification.
 *
 * Revision 1.5  1999/05/13 12:32:41  gerd
 * 	Data model change: There are now two tables, 'orig_links' and
 * 'rep_links' which have a common part and variant parts. Both are
 * handled by the same 'Table.t' derived class.
 *
 * Revision 1.4  1999/04/27 20:51:23  gerd
 * 	Change: "authors" and "links" are now tables known in Base. Note
 * that this has the consequence that these tables must be considered
 * as (coerced to) class Table.t, i.e. additional methods are no longer visible.
 *
 * Revision 1.3  1999/04/25 21:37:37  gerd
 * 	Bugfix: sort by date is now in descending order
 *
 * Revision 1.2  1999/04/25 18:22:16  gerd
 * 	Does not inherit from Layout anymore.
 *
 * Revision 1.1  1999/04/20 18:50:15  gerd
 * 	Moved from ../lib/ to this directory.
 *
 *
 * ======================================================================
 * OLD LOGS:
 *
 * Revision 1.3  1999/04/18 23:14:11  gerd
 * 	Major change: The Db tables have now typed attributes. The
 * attributes are represented as Field.field, grouped in records by
 * Record.record and stored in tables by File_base.t as before. A Scheme.scheme
 * is the signature of a table.
 * 	Change: Layout tries to encapsulate some of the layout properties
 * of tables.
 *
 * Revision 1.2  1999/03/14 01:08:40  gerd
 * 	Added: modification date as new record component
 * 	Added: by default, items are listed by modification date
 * 	Changed: passwords are stored as MD5 message digest
 * 	Added: can now change existing records field by field
 * 	Improved: design of the web pages
 *
 * Revision 1.1  1999/02/18 00:59:02  gerd
 * 	First version, has been tested to some extend.
 *
 * 
 *)
